searchData={"content_type":"text/markdown","items":[{"doc":"Evaluate user input expression.","ref":"ExPression.html","title":"ExPression","type":"module"},{"doc":"Evaluate expression.","ref":"ExPression.html#eval/2","title":"ExPression.eval/2","type":"function"},{"doc":"* `:bindings` - map variable names and values.\n* `:functions_module` - module with functions that will be accessible from expressions.","ref":"ExPression.html#eval/2-options","title":"Options - ExPression.eval/2","type":"function"},{"doc":"iex> eval(\"1 + 0.5\")\n    {:ok, 1.5}\n    iex> eval(\"div(x, y)\", bindings: %{\"x\" => 5, \"y\" => 2}, functions_module: Kernel)\n    {:ok, 2}\n    iex> eval(~s/{\"1\": \"en\", \"2\": \"fr\"}[str(int_code)]/, bindings: %{\"int_code\" => 1})\n    {:ok, \"en\"}\n    iex> eval(\"not true or false or 1 == 1\")\n    {:ok, true}\n    iex> eval(\"exit(self())\")\n    {:error, %ExPression.Error{name: \"UndefinedFunctionError\", message: \"Function 'self/0' was referenced, but was not defined\", data: %{function: :self}}}","ref":"ExPression.html#eval/2-examples","title":"Examples - ExPression.eval/2","type":"function"},{"doc":"Evaluate expression given in AST format.","ref":"ExPression.html#eval_ast/2","title":"ExPression.eval_ast/2","type":"function"},{"doc":"Parse expression in string format into AST format.\n\nThis can be used for optimizations: to parse expression once and evaluate AST many times.","ref":"ExPression.html#parse/1","title":"ExPression.parse/1","type":"function"},{"doc":"","ref":"ExPression.html#t:ast/0","title":"ExPression.ast/0","type":"type"},{"doc":"Error that occurred while parsing or evaluating expression.","ref":"ExPression.Error.html","title":"ExPression.Error","type":"module"},{"doc":"","ref":"ExPression.Error.html#new/3","title":"ExPression.Error.new/3","type":"function"},{"doc":"","ref":"ExPression.Error.html#t:t/0","title":"ExPression.Error.t/0","type":"type"},{"doc":"Evaluating AST","ref":"ExPression.Interpreter.html","title":"ExPression.Interpreter","type":"module"},{"doc":"","ref":"ExPression.Interpreter.html#eval/3","title":"ExPression.Interpreter.eval/3","type":"function"},{"doc":"Parsing expressions in strings format with convertion to AST format","ref":"ExPression.Parser.html","title":"ExPression.Parser","type":"module"},{"doc":"","ref":"ExPression.Parser.html#parse/1","title":"ExPression.Parser.parse/1","type":"function"},{"doc":"Expressions formal language grammar definition","ref":"ExPression.Parser.Grammar.html","title":"ExPression.Parser.Grammar","type":"module"},{"doc":"","ref":"ExPression.Parser.Grammar.html#peg/0","title":"ExPression.Parser.Grammar.peg/0","type":"function"},{"doc":"Standard library of functinos for ExPressions\n\nYou can override or add to standard functinos by providing a module to ExPression.","ref":"ExPression.StandardLib.html","title":"ExPression.StandardLib","type":"module"},{"doc":"Returns the absolute value of a number.","ref":"ExPression.StandardLib.html#abs/1","title":"ExPression.StandardLib.abs/1","type":"function"},{"doc":"Converts number or string to integer","ref":"ExPression.StandardLib.html#int/1","title":"ExPression.StandardLib.int/1","type":"function"},{"doc":"Return the length (the number of items) of an object.","ref":"ExPression.StandardLib.html#len/1","title":"ExPression.StandardLib.len/1","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#max/1","title":"ExPression.StandardLib.max/1","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#max/2","title":"ExPression.StandardLib.max/2","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#min/1","title":"ExPression.StandardLib.min/1","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#min/2","title":"ExPression.StandardLib.min/2","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#pow/2","title":"ExPression.StandardLib.pow/2","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#round/1","title":"ExPression.StandardLib.round/1","type":"function"},{"doc":"","ref":"ExPression.StandardLib.html#round/2","title":"ExPression.StandardLib.round/2","type":"function"},{"doc":"Converts term to it's string representation","ref":"ExPression.StandardLib.html#str/1","title":"ExPression.StandardLib.str/1","type":"function"},{"doc":"# ExPression\n[![Hex Version](https://img.shields.io/hexpm/v/ex_pression.svg)](https://hex.pm/packages/ex_pression)\n[![Docs](https://img.shields.io/badge/docs-hexpm-blue.svg)](https://hexdocs.pm/ex_pression)\n[![License](https://img.shields.io/hexpm/l/ex_pression.svg)](LICENSE)\n[![Coverage Status](https://coveralls.io/repos/github/balance-platform/ex_pression/badge.svg?branch=master)](https://coveralls.io/github/balance-platform/ex_pression?branch=master)\n\nEvaluate user input expressions.","ref":"readme.html","title":"ExPression","type":"extras"},{"doc":"The package can be installed by adding `ex_pression` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:ex_pression, \"~> 0.5.0\"}\n  ]\nend\n```","ref":"readme.html#installation","title":"Installation - ExPression","type":"extras"},{"doc":"1. **Safe** evaluation without access to other Elixir modules.\n```elixir\niex> ExPression.eval(\"exit(self())\")\n{:error, %ExPression.Error{name: \"UndefinedFunctionError\", message: \"Function 'self/0' was referenced, but was not defined\", data: %{function: :self}}}\n```\n\n2. Support for **JSON** syntax and data types.\n```elixir\niex> ExPression.eval(\"\"\"\n{\n  \"name\": \"ex_pression\",\n  \"deps\": [\"xpeg\"]\n}\n\"\"\")\n{:ok, %{\"name\" => \"ex_pression\", \"deps\" => [\"xpeg\"]}}\n```\n\n3. Familiar **python**-like operators and standard functions.\n```elixir\niex> ExPression.eval(~s/{\"1\": \"en\", \"2\": \"fr\"}[str(int_code)]/, bindings: %{\"int_code\" => 1})\n{:ok, \"en\"}\n```\n\n4. **Extend** expressions by providing Elixir module with functions that you want to use.\n```elixir\ndefmodule MyFunctions do\n  # use $ special symbol in expressions\n  def handle_special(\"$\", date_str), do: Date.from_iso8601!(date_str)\n  # Use diff function in expressions\n  def diff(date_1, date_2), do: Date.diff(date_1, date_2)\nend\n\niex> ExPression.eval(~s/diff($\"2023-02-02\", $\"2022-02-02\")/, functions_module: MyFunctions)\n{:ok, 365}\n```\n\nFull language description can be found in [FULL_DESCRIPTION.md](./FULL_DESCRIPTION.md)","ref":"readme.html#key-features","title":"Key features - ExPression","type":"extras"},{"doc":"String representation of expression is parsed into AST form. Parsing is done with PEG grammar parser [xpeg](https://github.com/zevv/xpeg). Grammar is defined in module `ExPression.Parser.Grammar`.\nAST interpretation logic is written in plain `Elixir` in module `ExPression.Interpreter`.","ref":"readme.html#implementation","title":"Implementation - ExPression","type":"extras"},{"doc":"Feel free to make a pull request. All contributions are appreciated!","ref":"readme.html#contribution","title":"Contribution - ExPression","type":"extras"},{"doc":"MIT License\n\nCopyright (c) 2024 Matvey Karpov\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.","ref":"license.html","title":"LICENSE","type":"extras"},{"doc":"# Full ExPressions language description","ref":"full_description.html","title":"Full ExPressions language description","type":"extras"},{"doc":"1. [Data types](#data_types)\n    1. [Numbers](#numbers)\n    2. [Booleans (and `null`)](#booleans)\n    3. [Strings](#strings)\n    4. [Arrays](#arrays)\n    5. [Objects](#objects)\n2. [Operators](#operators)\n    1. [Access operators](#access-operators)\n    2. [String operators](#string-operators)\n    3. [Arithmetic operators (and braces)](#arithmetic-operators)\n    4. [Logical operators](#logical-operators)\n3. [Function calls](#function-calls)\n4. [Special symbols](#special-symbols)\n5. [Variable bindings](#variable-bindings)\n6. [Extend expressions with new functions](#extend)\n7. [Standard library functions](#std)\n8. [Errors handling](#errors)","ref":"full_description.html#table-of-contents","title":"Table of contents - Full ExPressions language description","type":"extras"},{"doc":"Language has support for all JSON data types:\n#","ref":"full_description.html#data-types","title":"Data types - Full ExPressions language description","type":"extras"},{"doc":"Integers and floats:\n```elixir\niex> ExPression.eval(\"1\")\n{:ok, 1}\niex> ExPression.eval(\"1.01\")\n{:ok, 1.01}\niex> ExPression.eval(\"1e-3\")\n{:ok, 0.001}\n```\n\n#","ref":"full_description.html#numbers","title":"Numbers - Full ExPressions language description","type":"extras"},{"doc":"```elixir\niex> ExPression.eval(\"true\")\n{:ok, true}\niex> ExPression.eval(\"false\")\n{:ok, false}\niex> ExPression.eval(\"null\")\n{:ok, nil}\n```\n\n#","ref":"full_description.html#booleans-and-null","title":"Booleans (and `null`) - Full ExPressions language description","type":"extras"},{"doc":"```elixir\niex> ExPression.eval(~s(\"Hello, World!\"))\n{:ok, \"Hello, World!\"}\niex> ExPression.eval(~s(\"Привет, мир!\"))\n{:ok, \"Привет, мир!\"}\n```\n\n#","ref":"full_description.html#strings","title":"Strings - Full ExPressions language description","type":"extras"},{"doc":"```elixir\niex> ExPression.eval(\"[1, [2, 3]]\")\n{:ok, [1, [2, 3]]}\n```\n#","ref":"full_description.html#arrays","title":"Arrays - Full ExPressions language description","type":"extras"},{"doc":"```elixir\niex> ExPression.eval(~s({\"en\": \"England\", \"fr\": \"France\"}))\n{:ok, %{\"en\" => \"England\", \"fr\" => \"France\"}}\n```\n##","ref":"full_description.html#objects","title":"Objects - Full ExPressions language description","type":"extras"},{"doc":"Functions in expressions can return any of `Elixir` types:\n```elixir\ndefmodule MyFunctions do\n  def new_date(y, m, d), do: Date.new(y, m, d)\nend\n\niex> ExPression.eval(\"new_date(2024, 1, 1)\", functions_module: MyFunctions)\n{:ok, ~D[2022-01-30]}\n```","ref":"full_description.html#note","title":"Note: - Full ExPressions language description","type":"extras"},{"doc":"#","ref":"full_description.html#operators","title":"Operators - Full ExPressions language description","type":"extras"},{"doc":"1. Access object's field value using dot syntax:\n```elixir\niex> ExPression.eval(~s({\"en\": \"England\", \"fr\": \"France\"}.fr))\n{:ok, \"France\"}\n# dot operator also works with atom keys\niex> ExPression.eval(\"x.year\", bindings: %{\"x\" => ~D[2022-02-02]})\n{:ok, 2022}\n```\n2. Access object's field value with braces (in braces can be any expression):\n```elixir\niex> ExPression.eval(~s({\"en\": \"England\", \"fr\": \"France\"}[\"fr\"]))\n{:ok, \"France\"}\n```\n3. Access array's element by index:\n```elixir\niex> ExPression.eval(~s([1, 2, 3][0]))\n{:ok, 1}\n```\n\n#","ref":"full_description.html#access-operators","title":"Access operators - Full ExPressions language description","type":"extras"},{"doc":"`+` (concatenation):\n```elixir\niex> ExPression.eval(~s(\"abc\" + \"def\"))\n{:ok, \"abcdef\"}\n```\n\n#","ref":"full_description.html#string-operators","title":"String operators - Full ExPressions language description","type":"extras"},{"doc":"`+`, `-`, `*`, `/`:\n```elixir\niex> ExPression.eval(\"1 + 2 * (3 + 4) / 5\")\n{:ok, 3.8}\n```\nBraces are supported in general, not only for arithemtic operators, but here is an example.\n\n#","ref":"full_description.html#arithmetic-operators-and-braces","title":"Arithmetic operators (and braces) - Full ExPressions language description","type":"extras"},{"doc":"`and`, `or`, `not`, `==`, `!=`, `<`, `<=`, `>`, `>=`:\n```elixir\niex> ExPression.eval(\"not (1 == 2) or false\")\n{:ok, true}\n```\nOperators follow `python` semantics:\n```elixir\niex> ExPression.eval(~s(0 or \"\" or [] or [1, 2]))\n{:ok, [1, 2]}\niex> ExPression.eval(~s([] < [1, 2] and \"12\" < \"123\"))\n{:ok, true}\n```","ref":"full_description.html#logical-operators","title":"Logical operators - Full ExPressions language description","type":"extras"},{"doc":"Familiar syntax for function calls with [standard library of functions](#std).\n```elixir\niex> ExPression.eval(\"min(1, 2)\")\n{:ok, 1}\n```","ref":"full_description.html#function-calls","title":"Function calls - Full ExPressions language description","type":"extras"},{"doc":"Language supports special symbols (`$`) for most frequent operations in your domain field:\n```elixir\ndefmodule MyFunctions do\n  # use $ special symbol in expressions\n  def handle_special(\"$\", date_str), do: Date.from_iso8601!(date_str)\n  # Use diff function in expressions\n  def diff(date_1, date_2), do: Date.diff(date_1, date_2)\nend\n\niex> ExPression.eval(~s/diff($\"2023-02-02\", $\"2022-02-02\")/, functions_module: MyFunctions)\n{:ok, 365}\n```\nSpecial symbol can be followed by an identifier (`$i_am_special`) or a string (`$\"I am special!\"`).\nTo use special symbols, you have to provide `functions_module` and define a function `handle_special(symbol, value)`.\nReturn value of the function will be used in expression.","ref":"full_description.html#special-symbols","title":"Special symbols - Full ExPressions language description","type":"extras"},{"doc":"Pass variables to expressions using `:bindings` option:\n```elixir\niex> ExPression.eval(~s({\"en\": \"England\", \"fr\": \"France\"}[country_code]), bindings: %{\"country_code\" => \"en\"})\n{:ok, \"England\"}\n```","ref":"full_description.html#variable-bindings","title":"Variable bindings - Full ExPressions language description","type":"extras"},{"doc":"Define functions that you want to allow your user to use in expressions in some module, then pass this module with `:functions_module` option:\n```elixir\ndefmodule MyFunctions do\n  # Use my_sum function in expressions\n  def my_sum(a, b), do: a + b\nend\n\niex> ExPression.eval(\"my_sum(1, 2)\", functions_module: MyFunctions)\n{:ok, 3}\n```\nIn case your function and function from ExPression's standard library have the same name and arity, your function will be called.","ref":"full_description.html#extend-expressions-with-new-functions","title":"Extend expressions with new functions - Full ExPressions language description","type":"extras"},{"doc":"Functions have same names and signatures as `python`'s builtins.\n\n| Function | Argument types  | Description  |\n|---|---|---|\n| `len(term)` | `term` can be `string` or `array` or `object` | Return the length (the number of items) of an object |\n| `abs(term)` | `number` | Returns the absolute value of a number |\n| `str(term)` | `term` can be any of ExPression types | In case of `string` does nothing, in other cases - returns string representation of data |\n| `int(term)` | `term` can be `number` or `string` | truncates number to integer |\n| `round(number)` | `number` | rounds number to an integer |\n| `round(number, precision)` | `number` | rounds number with given `precision` |\n| `min(array)` | `array` | returns minimum of an array, raises `Enum.EmptyError` in case of empty array |\n| `min(a, b)` | `a` and `b` can be any comparable types | returns minimum of `a` and `b` |\n| `max(array)` | `array` | returns maximum of an array, raises `Enum.EmptyError` in case of empty array |\n| `max(a, b)` | `a` and `b` can be any comparable types | returns maximum of `a` and `b` |\n| `pow(base, exponent)` | `base` and `exponent` have to be `numbers` (integer or floating point) | returns `a` to the power of `b`. If both are integers - result is `integer`. Otherwise - `floating point` |","ref":"full_description.html#standard-library-functions","title":"Standard library functions - Full ExPressions language description","type":"extras"},{"doc":"In case of expected error, functions will return `{:error, error}` tuple, where `error` is a structure `%ExPression.Error{}`. In unexpected cases functions may raise an exception.\nHere is a list of types of possible errors:\n* `UndefinedVariableError` - when variable, that is used in expressions, was not found in `bindings` map.\n* `UndefinedFunctionError` - when function, that is used in expressions, was not found in `functions_module` and standard library.\n* `FunctionCallException` - when function, called in expression, raised an exception.\n* `BadOperationArgumentTypes` - invalid types of arguments for operation (for example, `1 + \"str\"`).\n* `SpecialWithoutModule` - when special symbol is used in expression, but module `:functions_module` was not provided.","ref":"full_description.html#errors-handling","title":"Errors handling - Full ExPressions language description","type":"extras"}]}